@@ -13,7 +13,7 @@ from PyQt5.QtChart import (
     QBarCategoryAxis,
 )
 from PyQt5.QtCore import QPointF, QLineF, Qt
-from PyQt5.QtGui import QPainter, QPainterPath, QColor, QPolygonF
+from PyQt5.QtGui import QPainter, QPainterPath, QColor, QPolygonF, QImage
 
 from agents.model import (
     get_quickstart_model,
@@ -32,66 +32,113 @@ from agents.model import (
 )
 
 
-class SimulationArea(QtWidgets.QWidget):
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self.__model = None
+class LineArea(QtWidgets.QWidget):
+    def __init__(self, model):
+        super().__init__()
+        self.model = model
+        self.lineimage = QImage(
+            model.width, model.height, QImage.Format_ARGB32
+        )
+        self.lineimage.fill(QColor(0, 0, 0, 0))
 
-    @property
-    def model(self):
-        return self.__model
+    def paintEvent(self, e):
+        linepainter = QtGui.QPainter(self.lineimage)
+        linepainter.setPen(QtCore.Qt.NoPen)
+        linepainter.setRenderHint(QtGui.QPainter.Antialiasing)
 
-    @model.setter
-    def model(self, model):
-        self.__model = model
-        self.setFixedSize(model.width, model.height)
-        self.enable_rendering = True
+        for agent in self.model.agents:
+            linepainter.setPen(
+                QColor(agent.color[0], agent.color[1], agent.color[2])
+            )
+            for line in agent.path.get_not_drawn():
+                linepainter.drawLine(
+                    QLineF(line[0][0], line[0][1], line[1][0], line[1][1])
+                )
+            agent.path.mark_as_drawn()
+        linepainter.end()
+
+        painter = QtGui.QPainter(self)
+        painter.setPen(QtCore.Qt.NoPen)
+        painter.setRenderHint(QtGui.QPainter.Antialiasing)
+        painter.drawImage(QPointF(0, 0), self.lineimage)
+        painter.end()
+
+
+class ShapeArea(QtWidgets.QWidget):
+    def __init__(self, model):
+        super().__init__()
+        self.model = model
+        self.shapeimage = QImage(
+            model.width, model.height, QImage.Format_ARGB32
+        )
+        self.shapeimage.fill(QColor(0, 0, 0, 0))
 
     def paintEvent(self, e):
         painter = QtGui.QPainter(self)
         painter.setPen(QtCore.Qt.NoPen)
         painter.setRenderHint(QtGui.QPainter.Antialiasing)
 
-        # Default to a black background
-        white = QtGui.QColor("black")
-        painter.setBrush(white)
-        painter.drawRect(
-            0, 0, painter.device().width(), painter.device().height()
+        for shape in self.model.get_shapes():
+            c = shape.color
+            painter.setBrush(QtGui.QColor(c[0], c[1], c[2]))
+            if type(shape) is EllipseStruct:
+                painter.drawEllipse(shape.x, shape.y, shape.w, shape.h)
+            elif type(shape) is RectStruct:
+                painter.drawRect(shape.x, shape.y, shape.w, shape.h)
+
+        painter.end()
+
+
+class TileArea(QtWidgets.QWidget):
+    def __init__(self, model):
+        super().__init__()
+        self.model = model
+        self.tileimage = QImage(
+            model.width, model.height, QImage.Format_ARGB32
         )
+        self.tileimage.fill(QColor(0, 0, 0))
+
+    def paintEvent(self, e):
+        tilepainter = QtGui.QPainter(self.tileimage)
+        tilepainter.setPen(QtCore.Qt.NoPen)
+        tilepainter.setRenderHint(QtGui.QPainter.Antialiasing)
 
         if self.model:
-            # Draw tiles
-            for tile in self.model.tiles:
-                self.paintTile(painter, tile)
-
-            # Draw shapes
-            for shape in self.model.get_shapes():
-                c = shape.color
-                painter.setBrush(QtGui.QColor(c[0], c[1], c[2]))
-                if type(shape) is EllipseStruct:
-                    painter.drawEllipse(shape.x, shape.y, shape.w, shape.h)
-                elif type(shape) is RectStruct:
-                    painter.drawRect(shape.x, shape.y, shape.w, shape.h)
-
-            # Draw lines
-            for agent in self.model.agents:
-                painter.setPen(
-                    QColor(agent.color[0], agent.color[1], agent.color[2])
+            for tile in self.model._vu_tiles:
+                r, g, b = tile.color
+                color = QtGui.QColor(r, g, b)
+                tilepainter.setBrush(color)
+                x = self.model.tile_size * tile.x
+                y = self.model.tile_size * tile.y
+                tilepainter.drawRect(
+                    x, y, self.model.tile_size, self.model.tile_size
                 )
-                # Pretty terrible code here
-                # Replaces line-tuples in agent.__paths with QLineF's
-                for path in agent.get_paths():
-                    for i in range(len(path)):
-                        if type(path[i]) is tuple:
-                            path[i] = QLineF(
-                                path[i][0][0],
-                                path[i][0][1],
-                                path[i][1][0],
-                                path[i][1][1],
-                            )
-                    painter.drawLines(path)
+            self.model._vu_tiles = []
+        tilepainter.end()
 
-            # Draw agents
+        painter = QtGui.QPainter(self)
+        painter.setPen(QtCore.Qt.NoPen)
+        painter.setRenderHint(QtGui.QPainter.Antialiasing)
+        painter.drawImage(QPointF(0, 0), self.tileimage)
+        painter.end()
+
+
+class AgentArea(QtWidgets.QWidget):
+    def __init__(self, model):
+        super().__init__()
+        self.model = model
+        self.agentimage = QImage(
+            model.width, model.height, QImage.Format_ARGB32
+        )
+        self.agentimage.fill(QColor(0, 0, 0, 0))
+
+    def paintEvent(self, e):
+        painter = QtGui.QPainter(self)
+        painter.setPen(QtCore.Qt.NoPen)
+        painter.setRenderHint(QtGui.QPainter.Antialiasing)
+
+        if self.model:
+            self.agentimage.fill(QColor(0, 0, 0, 0))
             select = None
             for agent in self.model.agents:
                 self.paintAgent(painter, agent)
@@ -109,6 +156,7 @@ class SimulationArea(QtWidgets.QWidget):
                 )
                 painter.setBrush(QColor(0, 0, 0, 150))
                 painter.drawPath(path)
+
         painter.end()
 
     def paintAgent(self, painter, agent):
@@ -179,13 +227,21 @@ class SimulationArea(QtWidgets.QWidget):
             ]
             painter.drawPolygon(QPolygonF(point_list))
 
-    def paintTile(self, painter, tile):
-        r, g, b = tile.color
-        color = QtGui.QColor(r, g, b)
-        painter.setBrush(color)
-        x = self.model.tile_size * tile.x
-        y = self.model.tile_size * tile.y
-        painter.drawRect(x, y, self.model.tile_size, self.model.tile_size)
+
+class SimulationArea(QtWidgets.QStackedLayout):
+    def __init__(self, wrapper, model):
+        super().__init__(wrapper)
+        self.addWidget(AgentArea(model))
+        self.addWidget(LineArea(model))
+        self.addWidget(ShapeArea(model))
+        self.addWidget(TileArea(model))
+        self.setStackingMode(1)
+        self.enable_rendering = True
+        self.__model = model
+
+    def draw(self):
+        for i in range(self.count()):
+            self.widget(i).update()
 
     def mousePressEvent(self, e):
         x = e.localPos().x()
@@ -567,7 +623,12 @@ class Application:
 
         # Box for left side (simulation area + controllers)
         self.left_box = QtWidgets.QVBoxLayout()
-        self.horizontal_divider.addLayout(self.left_box)
+        self.constrain_widget = QtWidgets.QWidget()
+        self.constrain_widget.setLayout(self.left_box)
+        self.constrain_widget.setMaximumSize(
+            self.model.width, self.constrain_widget.maximumHeight()
+        )
+        self.horizontal_divider.addWidget(self.constrain_widget)
         self.controllers = []
 
         # Box for right side (plots)
@@ -578,9 +639,10 @@ class Application:
         # self.right_box.addStretch(1)
 
         # Simulation area
-        self.simulation_area = SimulationArea()
-        self.simulation_area.model = self.model
-        self.left_box.addWidget(self.simulation_area)
+        self.area_wrapper = QtWidgets.QWidget()
+        self.simulation_area = SimulationArea(self.area_wrapper, self.model)
+        self.area_wrapper.setFixedSize(self.model.width, self.model.height)
+        self.left_box.addWidget(self.area_wrapper)
 
         # Controller box (bottom left)
         self.controller_box = QtWidgets.QVBoxLayout()
@@ -612,7 +674,7 @@ class Application:
 
     def update_graphics(self):
         if self.simulation_area.enable_rendering:
-            self.simulation_area.update()
+            self.simulation_area.draw()
         for p in self.model.plots:
             p.redraw()
 
